# -*- coding: utf-8 -*-
import os
import pytest
from pyramid.testing import DummyRequest, testConfig
from pyramid.httpexceptions import HTTPOk, HTTPFound, HTTPInternalServerError
from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
from pyramid_saml import SAML, SAML_ERROR, SAML_NAME_ID, SAML_SESSION_INDEX, \
    SAML_USER_DATA, _load_settings, _load_settings_from_file, \
    _load_settings_from_url


def __get_request():
    request = DummyRequest()
    request.scheme = 'https'
    request.server_port = 80
    return request


@pytest.mark.parametrize('settings,lowercase,scheme,https', [
    (
        {
            'index_route_name': 'index',
            'saml_path': os.path.dirname(__file__)
        },
        False,
        'https',
        'on'
    ),
    (
        {
            'index_route_name': 'index',
            'saml_path': os.path.dirname(__file__),
            'lowercase_urlencoding': True
        },
        True,
        'http',
        'off'
    )
])
def test_init(settings, lowercase, scheme, https):
    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings(config)
        config.add_route('index', '/')
        request = __get_request()
        request.scheme = scheme
        saml = SAML(request)

        assert isinstance(saml, SAML)
        assert saml._request == request
        assert saml._settings == settings
        assert saml._saml_request['lowercase_urlencoding'] == lowercase
        assert saml._saml_request['https'] == https
        assert isinstance(saml.auth, OneLogin_Saml2_Auth)


def test_metadata():

    def validate_metadata(metadata):
        return []

    settings = {
        'index_route_name': 'index',
        'saml_path': os.path.dirname(__file__)
    }
    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings(config)
        config.add_route('index', '/')
        request = __get_request()
        saml = SAML(request)
        settings = saml.auth.get_settings()
        settings.validate_metadata = validate_metadata
        response = saml.metadata()

        assert isinstance(response, HTTPOk)
        assert response.headers['Content-Type'] == 'text/xml'


def test_metadata_error():

    def validate_metadata(metadata):
        return ['error1']

    settings = {
        'index_route_name': 'index',
        'saml_path': os.path.dirname(__file__)
    }
    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings(config)
        config.add_route('index', '/')
        request = __get_request()
        saml = SAML(request)
        settings = saml.auth.get_settings()
        settings.validate_metadata = validate_metadata

        with pytest.raises(HTTPInternalServerError) as e:
            saml.metadata()
            assert '{}'.format(e) == 'error1'


@pytest.mark.parametrize('redirect', [
    None,
    'https://example.com/foo'
])
def test_sso(redirect):

    def login(return_to=None):
        return return_to

    settings = {
        'index_route_name': 'index',
        'saml_path': os.path.dirname(__file__)
    }
    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings(config)
        config.add_route('index', '/')
        request = __get_request()
        if redirect is not None:
            request.params['redirect'] = redirect
        saml = SAML(request)
        saml.auth.login = login
        response = saml.sso()

        assert isinstance(response, HTTPFound)
        if redirect is not None:
            assert response.location == redirect
        else:
            assert response.location == 'http://example.com/'


def test_slo():

    def logout(name_id=None, session_index=None):
        assert name_id == 'foo'
        assert session_index == 'bar'
        return 'http://example.com/'

    settings = {
        'index_route_name': 'index',
        'saml_path': os.path.dirname(__file__)
    }
    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings(config)
        config.add_route('index', '/')
        request = __get_request()
        request.session[SAML_NAME_ID] = 'foo'
        request.session[SAML_SESSION_INDEX] = 'bar'
        saml = SAML(request)
        saml.auth.logout = logout
        response = saml.slo()

        assert isinstance(response, HTTPFound)
        assert response.location == 'http://example.com/'


@pytest.mark.parametrize('errors,relay_state', [
    (['error1'], None),
    ([], None),
    ([], 'http://example.com/foo')
])
def test_acs(errors, relay_state):

    def process_response():
        pass

    def get_errors():
        return errors

    def get_last_error_reason():
        if len(errors) > 0:
            return errors[-1]
        return None

    def get_attributes():
        return {
            'foo': 'bar'
        }

    def get_nameid():
        return 'NameId'

    def get_session_index():
        return 'SessionIndex'

    def redirect_to(url):
        return url

    settings = {
        'index_route_name': 'index',
        'saml_path': os.path.dirname(__file__)
    }
    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings(config)
        config.add_route('index', '/')
        request = __get_request()
        if relay_state is not None:
            request.POST['RelayState'] = relay_state
        saml = SAML(request)
        saml.auth.process_response = process_response
        saml.auth.get_errors = get_errors
        saml.auth.get_last_error_reason = get_last_error_reason
        saml.auth.get_attributes = get_attributes
        saml.auth.get_nameid = get_nameid
        saml.auth.get_session_index = get_session_index
        saml.auth.redirect_to = redirect_to
        response = saml.acs()

        assert isinstance(response, HTTPFound)
        if len(errors) > 0:
            assert saml._session[SAML_ERROR] == errors[-1]
            assert SAML_USER_DATA not in saml._session
            assert SAML_NAME_ID not in saml._session
            assert SAML_SESSION_INDEX not in saml._session
        else:
            assert saml._session[SAML_ERROR] is None
            assert saml._session[SAML_USER_DATA] == {
                'foo': 'bar'
            }
            assert saml._session[SAML_NAME_ID] == 'NameId'
            assert saml._session[SAML_SESSION_INDEX] == 'SessionIndex'
        if relay_state is None:
            assert response.location == 'http://example.com/'
        else:
            assert response.location == relay_state


@pytest.mark.parametrize('url,errors',  [
    (None, ['error1']),
    ('https://example.com/foo', [])
])
def test_sls(url, errors):

    def process_slo(delete_session_cb=None):
        delete_session_cb()
        return url

    def get_errors():
        return errors

    def get_last_error_reason():
        if len(errors) > 0:
            return errors[-1]
        return None

    settings = {
        'index_route_name': 'index',
        'saml_path': os.path.dirname(__file__)
    }
    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings(config)
        config.add_route('index', '/')
        request = __get_request()
        saml = SAML(request)
        saml.auth.process_slo = process_slo
        saml.auth.get_errors = get_errors
        saml.auth.get_last_error_reason = get_last_error_reason
        saml._session[SAML_NAME_ID] = 'foo'
        response = saml.sls()

        assert isinstance(response, HTTPFound)
        if url is None:
            assert response.location == 'http://example.com/'
        else:
            assert response.location == url
        if len(errors) > 0:
            assert saml._session[SAML_ERROR] == errors[-1]
        else:
            assert SAML_ERROR not in saml._session
        assert SAML_NAME_ID not in saml._session


def test_load_settings_from_file():
    settings = {
        'saml_path': os.path.dirname(__file__)
    }
    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings_from_file(config)
        cfg = config.get_settings()['pyramid_saml']
        assert cfg['debug']
        assert cfg['sp']['entityId'] == 'https://example.com/metadata/'
        assert cfg['idp']['entityId'] == \
            'https://app.onelogin.com/saml/metadata/0'
        assert cfg['organization']['en-US']['name'] == 'sp_test'


@pytest.mark.parametrize('settings,env_url,env_id,num', [
    ({
        'idp': {
            'entityId': 'initial'
        }
    }, None, None, 1),
    ({
        'metadata_url': 'http://example.com/metadata2'
    }, None, None, 2),
    ({
        'idp': {
            'entityId': 'initial'
        }
    }, 'http://example.com/metadata3', None, 3),
    ({
        'idp': {
            'entityId': 'initial'
        }
    }, 'http://example.com/metadata4', 'from_env', 4)
])
def test_load_settings_from_url(settings, env_url, env_id, num):

    def parse_remote(metadata_url, entity_id=None):
        result = {
            'idp': {
                'url': metadata_url
            }
        }
        if entity_id is not None:
            result['idp']['entityId'] = entity_id
        return result

    OneLogin_Saml2_IdPMetadataParser.parse_remote = parse_remote

    if env_url is not None:
        os.environ['PYRAMID_SAML_METADATA_URL'] = env_url

    if env_id is not None:
        os.environ['PYRAMID_SAML_IDP_ENTITY_ID'] = env_id

    with testConfig(settings={'pyramid_saml': settings}) as config:
        _load_settings_from_url(config)
        idp = config.get_settings()['pyramid_saml']['idp']
        if num == 1:
            assert idp['entityId'] == 'initial'
            assert 'url' not in idp
        elif num == 2:
            assert 'entityId' not in idp
            assert idp['url'] == 'http://example.com/metadata2'
        elif num == 3:
            assert idp['entityId'] == 'initial'
            assert idp['url'] == 'http://example.com/metadata3'
        elif num == 4:
            assert idp['entityId'] == 'from_env'
            assert idp['url'] == 'http://example.com/metadata4'


@pytest.mark.parametrize('settings', [
    {},
    {
        'pyramid_saml': {}
    }
])
def test_load_settings_fail(settings):
    with testConfig(settings=settings) as config:
        with pytest.raises(AssertionError):
            _load_settings(config)


@pytest.mark.parametrize('settings,env', [
    ({}, {
        'PYRAMID_SAML_PATH': os.path.dirname(__file__)
    }),
    ({
        'pyramd_saml': {}
    }, {
        'PYRAMID_SAML_PATH': os.path.dirname(__file__)
    }),
    ({
        'pyramd_saml': {
            'saml_path': os.path.dirname(__file__)
        }
    }, {})
])
def test_load_settings(settings, env):
    os.environ.update(env)
    with testConfig(settings=settings) as config:
        _load_settings(config)
        cfg = config.get_settings()['pyramid_saml']
        assert cfg['debug']
        assert cfg['sp']['entityId'] == 'https://example.com/metadata/'



# -*- coding: utf-8 -*-
import os
import logging
import json
from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.utils import OneLogin_Saml2_Utils
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
from pyramid.httpexceptions import HTTPFound, HTTPOk, HTTPInternalServerError


__VERSION__ = '0.0.1'


log = logging.getLogger('pyramid_saml')
"""logging.Logger: Logger instance for log output."""


SAML_NAME_ID = 'samlNameId'
"""str: Key of the NameId attribute."""

SAML_USER_DATA = 'samlUserData'
"""str: Key of the UserData attribute."""

SAML_SESSION_INDEX = 'samlSessionIndex'
"""str: Key of the SessionIndex attribute."""

SAML_ERROR = 'samlError'
"""str: Key of the Error attribute."""


class SAML(object):
    def __init__(self, request):
        """The SAML object, providing the view methods.

        Args:
            request (pyramid.request.Request): The Pyramid request object.
        """

        self._request = request
        """pyramid.request.Request: The current request object."""

        self._session = self._request.session
        """dict: The current request's session instance."""

        self._settings = self._request.registry.settings.get('pyramid_saml')
        """dict: The parsed `pyramid_saml` settings."""

        # Lowercase URL encoding is needed for ADFS
        lowercase = self._settings.get('lowercase_urlencoding') or False

        saml_path = self._settings.get('saml_path')
        self._saml_path = os.path.dirname(__file__) \
            if saml_path is None \
            else saml_path
        """str: Directory containing the settings.json and certificates."""

        # Name of a default route as fallback for redirects
        self._index_url = self._request.route_url(
            self._settings.get('index_route_name')
        )
        """str: Index route URL."""

        self._saml_request = {
            'https': 'on' if request.scheme == 'https' else 'off',
            'http_host': request.host,
            'server_port': request.server_port,
            'script_name': request.path,
            'get_data': request.GET.copy(),
            'lowercase_urlencoding': lowercase,
            'post_data': request.POST.copy(),
        }
        """dict: The SAML request configuration."""

        self.auth = OneLogin_Saml2_Auth(
            self._saml_request,
            self._settings,
            custom_base_path=self._saml_path
        )
        """onelogin.saml2.auth.OneLogin_Saml2_Auth: The current auth
                object.
        """

    def sso(self):
        """Performs the login request.

        Returns:
            pyramid.httpexceptions.HTTPFound: Redirect to the IdP.
        """
        if 'redirect' in self._request.params:
            return_to = self._request.params.get('redirect')
        else:
            return_to = self._index_url
        return HTTPFound(self.auth.login(return_to))

    def slo(self):
        """Performs the logout request.

        Returns:
            pyramid.httpexceptions.HTTPFound: Redirect to the IdP.
        """
        name_id = None
        session_index = None
        if SAML_NAME_ID in self._session:
            name_id = self._session[SAML_NAME_ID]
        if SAML_SESSION_INDEX in self._session:
            session_index = self._session[SAML_SESSION_INDEX]
        return HTTPFound(self.auth.logout(
            name_id=name_id,
            session_index=session_index
        ))

    def acs(self):
        """Process the IdP's login response.

        Returns:
            pyramid.httpexceptions.HTTPFound: Redirect to specified URL or
                configured index.
        """
        self.auth.process_response()
        errors = self.auth.get_errors()
        if len(errors) == 0:
            self._session[SAML_USER_DATA] = self.auth.get_attributes()
            self._session[SAML_NAME_ID] = self.auth.get_nameid()
            self._session[SAML_SESSION_INDEX] = self.auth.get_session_index()
            self._session[SAML_ERROR] = None
            self_url = OneLogin_Saml2_Utils.get_self_url(self._saml_request)
            if 'RelayState' in self._request.POST \
                    and self_url != self._request.POST['RelayState']:
                url = self.auth.redirect_to(self._request.POST['RelayState'])
                return HTTPFound(url)
        else:
            msg = self.auth.get_last_error_reason()
            log.error(msg)
            self._session[SAML_ERROR] = msg
        return HTTPFound(self._index_url)

    def sls(self):
        """Process the IdP's logout response.

        Returns:
            pyramid.httpexceptions.HTTPFound: Redirect to specified URL or
                configured index.
        """

        def delete_session_cb():
            """Callback function to clear the current session."""
            self._session.clear()

        url = self.auth.process_slo(delete_session_cb=delete_session_cb)
        errors = self.auth.get_errors()
        if len(errors) > 0:
            log.error(self.auth.get_last_error_reason())
            self._session[SAML_ERROR] = self.auth.get_last_error_reason()
        if url is not None:
            return HTTPFound(url)
        else:
            return HTTPFound(self._index_url)

    def metadata(self):
        """Returns the SP's metadata.

        Returns:
            pyramid.httpexceptions.HTTPOk: The rendered metadata.

        Raises:
            pyramid.httpexceptions.HTTPInternalServerError: The metadata
                error.
        """
        settings = self.auth.get_settings()
        metadata = settings.get_sp_metadata()
        errors = settings.validate_metadata(metadata)

        if len(errors) > 0:
            raise HTTPInternalServerError(', '.join(errors))

        return HTTPOk(body=metadata, headers={'Content-Type': 'text/xml'})


def _load_settings_from_file(config):
    """Parse configuration from `settings.json` and `advanced_settings.json`.

    Args:
        config (pyramid.config.Configurator): The configuration of the
            existing Pyramid application.
    """

    saml_path = config.get_settings().get('pyramid_saml').get('saml_path')

    with open(os.path.join(saml_path, 'settings.json')) as f:
        config.get_settings().get('pyramid_saml').update(
            json.loads(f.read())
        )

    if os.path.isfile(os.path.join(saml_path, 'advanced_settings.json')):
        with open(os.path.join(saml_path, 'advanced_settings.json')) as f:
            config.get_settings().get('pyramid_saml').update(
                json.loads(f.read())
            )


def _load_settings_from_url(config):
    """Parse configuration from IdP metadata.

    Args:
        config (pyramid.config.Configurator): The configuration of the
            existing Pyramid application.
    """

    settings = config.get_settings().get('pyramid_saml')

    metadata_url = entity_id = None
    if 'metadata_url' in settings:
        metadata_url = settings.get('metadata_url')
    if 'PYRAMID_SAML_METADATA_URL' in os.environ:
        metadata_url = os.environ.get('PYRAMID_SAML_METADATA_URL')
    if 'idp' in settings and 'entityId' in settings.get('idp'):
        entity_id = settings.get('idp').get('entityId')
    if 'PYRAMID_SAML_IDP_ENTITY_ID' in os.environ:
        entity_id = os.environ.get('PYRAMID_SAML_IDP_ENTITY_ID')

    if metadata_url is not None:
        metadata = OneLogin_Saml2_IdPMetadataParser.parse_remote(
            metadata_url,
            entity_id=entity_id
        )
        log.debug('Parsed IDP metadata:\n{0}'.format(
            json.dumps(metadata, indent=4)
        ))
        if 'idp' in settings:
            config.get_settings()['pyramid_saml']['idp'].update(
                metadata['idp']
            )
        else:
            config.get_settings()['pyramid_saml']['idp'] = metadata['idp']


def _load_settings(config):
    """Parse and validate SAML configuration.

    Args:
        config (pyramid.config.Configurator): The configuration of the
            existing Pyramid application.
    """

    if 'PYRAMID_SAML_PATH' in os.environ:
        if 'pyramid_saml' not in config.get_settings():
            config.get_settings().update({
                'pyramid_saml': {}
            })
        config.get_settings().get('pyramid_saml').update({
            'saml_path': os.environ.get('PYRAMID_SAML_PATH')
        })

    if 'pyramid_saml' not in config.get_settings():
        msg = 'Missing \'pyramid_saml\' configuration in settings'
        raise AssertionError(msg)
    else:
        settings = config.get_settings().get('pyramid_saml')

    if 'saml_path' not in settings:
        msg = 'Missing \'saml_path\' in settings'
        raise AssertionError(msg)
    else:
        saml_path = settings.get('saml_path')
        log.debug('SAML path: {0}'.format(saml_path))

    _load_settings_from_file(config)
    _load_settings_from_url(config)

    log.debug('SAML config:\n{0}'.format(json.dumps(
        config.get_settings().get('pyramid_saml'),
        indent=4
    )))


def includeme(config):  # pragma: no cover
    """Include this library in an existing Pyramid application.

    Args:
        config (pyramid.config.Configurator): The configuration of the
            existing Pyramid application.
    """

    # Set log level through env variable
    log_level = '{0}'.format(os.environ.get('PYRAMID_SAML_LOG_LEVEL')).lower()
    if log_level == 'error':
        log.setLevel(logging.ERROR)
    elif log_level == 'warning':
        log.setLevel(logging.WARNING)
    elif log_level == 'debug':
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    config.add_route('pyramid_saml_sso', '/sso')
    config.add_view(
        SAML,
        attr='sso',
        route_name='pyramid_saml_sso'
    )

    config.add_route('pyramid_saml_slo', '/slo')
    config.add_view(
        SAML,
        attr='slo',
        route_name='pyramid_saml_slo'
    )

    config.add_route('pyramid_saml_acs', '/acs')
    config.add_view(
        SAML,
        attr='acs',
        route_name='pyramid_saml_acs'
    )

    config.add_route('pyramid_saml_sls', '/sls')
    config.add_view(
        SAML,
        attr='sls',
        route_name='pyramid_saml_sls'
    )

    config.add_route('pyramid_saml_metadata', '/metadata')
    config.add_view(
        SAML,
        attr='metadata',
        route_name='pyramid_saml_metadata'
    )

    # Load SAML settings and append them to the application settings
    _load_settings(config)



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

requires = [
    'pyramid',
    'python3-saml'
]

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md')) as f:
    long_description = f.read()

with open(os.path.join(here, 'CHANGELOG.md')) as f:
    long_description += '\n\n'
    long_description += f.read()

setup(
    name='pyramid_saml',
    version='0.0.1',
    description='Pyramid SAML',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Karsten Deininger',
    author_email='karsten.deininger@bl.ch',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Pyramid',
        'Topic :: Internet :: WWW/HTTP'
    ],
    url='https://gitlab.com/geo-bl-ch/pyramid_saml',
    keywords='web pyramid saml',
    install_requires=requires,
    packages=find_packages(exclude=['demo', 'test*']),
    include_package_data=True
)
